==> rtp_qsndqs-0.1.tgz: rtp_qsndqs/prolog/rtp_qs.pl <==
% -*- prolog -*-
/** <module> Real-time Persistent Queues

Offers a Prolog implementation of queues with constant time basic
operations and full version reusability.

See: Bouma, Gerlof. 2012. Real-time Persistent Queues and Deques with
  Logic Variables (Declarative Pearl). In Schrijvers and Thiemann
  (Eds), Functional and Logic Programming, Proceedings of the 11th
  International Symposium FLOPS 2012, LNCS 7294, pp. 62-72,
  Heidelberg: Springer.

Time guarantees and determinism indications apply to the cases in
which the predicate signature is respected.

In the signatures, +Q:queue means that Q must be sufficiently
instantiated, that is, it is a valid queue representing term as
constructed by a queue outputting predicate. Due to the design of the
datastructure, these need not be ground.

@author Gerlof Bouma, Uni Gothenburg
@tbd Loading together with rtp_dqs.pl somehow doesn't work.
*/

:- module(rtp_qs,
        [ empty_queue/1,         % creates/tests for the empty queue
          inject_queue/3,        % add to back of queue
          pop_queue/3,           % remove from front of queue
          push_queue/3,          % add to front of queue -- it is really
	                         %   an output-restricted queue!
          poplist_queue/3,       % add a list of elements to
          pushlist_queue/3,      %   to front or back or 
          injectlist_queue/3     %   remove from the front
      ]).

:- use_module(library(type_check)).

:- type queue(T) ---> q(list(T),list(T),state(T)).
:- type state(T) ---> ready
                      % Rs_revs
                      ; wait(list(T))
		      %      St
                      ; delay(state(T))
                      %     Fs       Fs_open Rs     Rs_revs
                      ; rot(list(T),list(T),list(T),list(list(T))).


%% empty_queue(?Q:queue) is semidet.
% 
% Holds of the empty queue.
%
:- pred empty_queue(queue(_T)).
%
empty_queue(q([],[],ready)).


%% pop_queue(+Q_old:queue,?El,-Q_new:queue) is semidet.
%
% Removes ?El from the front of Q_old to give Q_new. 
%
% May fail if ?El cannot be unified with the front element of Q_old
% or if Q_old is empty.
%
:- pred pop_queue(queue(T),T,  queue(T)).
%
pop_queue(q([F|Fs],Rs,St),F,Q1):-
	make_q(St,Fs,Rs,Q1).


%% inject_deque(?El,+Q_old:queue,-Q_new:queue) is det.
%
% Adds El to the back of Q_old to give Q_new.
%
:- pred inject_queue(T,  queue(T),queue(T)).
%
inject_queue(R,q(Fs,Rs,St),Q1):-
	make_q(St,Fs,[R|Rs],Q1).


%% push_queue(?El,+DQ_old:queue,-DQ_new:queue) is det.
%
% Adds El to the front of DQ_old to give DQ_new.
%
% ( This extension of the version presented in the paper makes the
% datastructure an output-restricted double-ended queue. )
%
:- pred push_queue(T,  queue(T),queue(T)).
%
push_queue(F,q(Fs,Rs,St),q([F|Fs],Rs,delay(St))).


%% poplist_queue(+Q_old:queue,?Els:list,-Q_new:queue) is nondet.
%
% Pops elements from Q_old onto Els to give Q_new. See pop_queue/3.
%
% Gives a.o. quick (and dirty?)
% ways of implementing queue-to-list 
% ( =|poplist_queue(Q_old,Els,Q_new), empty(Q_new)|= ) and  `take N'
% ( =|length(N,Els), poplist(Q_old,Els,Q_new)|= ). 
%
:- pred pops_queue(queue(T),list(T),queue(T)).
%
poplist_queue(Q,[],Q).
poplist_queue(Q,[El|Els],Q2):-
	pop_queue(Q,El,Q1),
	poplist_queue(Q1,Els,Q2).


%% injectlist_queue(+Els:list,+Q_old:queue,-Q_new:queue) is det.
%
% Injects elements of Els into Q_old to give Q_new. 
%
:- pred injects_queue(list(T),queue(T),queue(T)).
%
injectlist_queue([],Q,Q).
injectlist_queue([El|Els],Q,Q2):-
	inject_queue(El,Q,Q1),
	injectlist_queue(Els,Q1,Q2).


%% pushlist_queue(+Els:list,+Q_old:queue,-Q_new:queue) is det.
%
% Pushs elements of Els onto Q_old to give Q_new. 
%
:- pred pushs_queue(list(T),queue(T),queue(T)).
%
pushlist_queue([],Q,Q).
pushlist_queue([El|Els],Q,Q2):-
	push_queue(El,Q,Q1),
	pushlist_queue(Els,Q1,Q2).


% Aux predicates. See the paper for explanation. Compared to the
% paper, these are longer, `unrolled' versions that do not use the
% second-order predicate call/2.
%

% make_q/4      +State, +Fs     +Rs     -Q1
:- pred make_q(state(T),list(T),list(T),queue(T)).
%
make_q(ready,Fs,Rs,q(Fs1,[],St1)):-
	rot(Fs,Fs1,Rs,[[]|_],St1).
make_q(wait(A),Fs,Rs,q(Fs,Rs,St1)):-
    wait(A,St1).
make_q(delay(St1),Fs,Rs,q(Fs,Rs,St1)).         
make_q(rot(A1,A2,A3,A4),Fs,Rs,q(Fs,Rs,St1)):-
	rot(A1,A2,A3,A4,St1).


% wait/2     +Rs_rev -State
:- pred(wait(list(T),state(T))).
wait([],ready).
wait([_|Rs_rev],wait(Rs_rev)).


% rot/5     +Fs     ?Fs_tail +Rs     ?Rs_revs      -State
:- pred(rot(list(T),list(T), list(T),list(list(T)),state(T))).
%
rot([],[R|Rs_rev],[R],[Rs_rev],wait(Rs_rev)).
rot([F|Fs],[F|Fs_tail],[R|Rs],[Rs_rev|Rs_revs],rot(Fs,Fs_tail,Rs,Rs_revs)):-
	Rs_revs = [[R|Rs_rev]|_].

==> rtp_qsndqs-0.3.tgz: rtp_qsndqs/prolog/rtp_qs.pl <==
% -*- prolog -*-
:- module(rtp_qs,
        [ empty_queue/1,         % creates/tests for the empty queue
          inject_queue/3,        % add to back of queue
          pop_queue/3,           % remove from front of queue
          push_queue/3,          % add to front of queue -- it is really
	                         %   an output-restricted deque!
          poplist_queue/3,       % add a list of elements to
          pushlist_queue/3,      %   to front or back or 
          injectlist_queue/3     %   remove from the front
      ]).

:- use_module(library(type_check)).

/** <module> Real-time Persistent Queues

Offers a Prolog implementation of queues with constant time basic
operations and full version reusability. The implementation is based
on Bouma (2012), referenced below. Time guarantees and determinism
indications apply to the cases in which the predicate signature is
respected.

In the signatures, +Q:queue means that Q is an input argument and must
be _sufficiently_ instantiated, that is, it is a valid queue
representing term as constructed by a queue outputting predicate. Due
to the design of the datastructure, these need not be ground.

@author Gerlof Bouma, Uni Gothenburg
@see Bouma, Gerlof. 2012. Real-time Persistent Queues and Deques with
  Logic Variables (Declarative Pearl). In Schrijvers and Thiemann
  (Eds), Functional and Logic Programming, Proceedings of the 11th
  International Symposium FLOPS 2012, LNCS 7294, pp. 62-72,
  Heidelberg: Springer.
@see http://spraakbanken.gu.se/personal/gerlof/home
@bug Loading together with rtp_dqs.pl somehow doesn't work.
*/

:- type queue(T) ---> q(list(T),list(T),state(T)).
:- type state(T) ---> ready
                      % Rs_revs
                      ; wait(list(T))
		      %      St
                      ; delay(state(T))
                      %     Fs       Fs_open Rs     Rs_revs
                      ; rot(list(T),list(T),list(T),list(list(T))).


%% empty_queue(?Q:queue) is semidet.
% 
% Holds of the empty queue.
%
:- pred empty_queue(queue(_T)).
%
empty_queue(q([],[],ready)).


%% pop_queue(+Q_old:queue,?El,-Q_new:queue) is semidet.
%
% Removes ?El from the front of Q_old to give Q_new. 
%
% May fail if ?El cannot be unified with the front element of Q_old
% or if Q_old is empty.
%
:- pred pop_queue(queue(T),T,  queue(T)).
%
pop_queue(q([F|Fs],Rs,St),F,Q1):-
	make_q(St,Fs,Rs,Q1).


%% inject_queue(?El,+Q_old:queue,-Q_new:queue) is det.
%
% Adds El to the back of Q_old to give Q_new.
%
:- pred inject_queue(T,  queue(T),queue(T)).
%
inject_queue(R,q(Fs,Rs,St),Q1):-
	make_q(St,Fs,[R|Rs],Q1).


%% push_queue(?El,+DQ_old:queue,-DQ_new:queue) is det.
%
% Adds El to the front of DQ_old to give DQ_new.
%
% ( This extension of the version presented in the paper makes the
% datastructure an output-restricted double-ended queue. )
%
:- pred push_queue(T,  queue(T),queue(T)).
%
push_queue(F,q(Fs,Rs,St),q([F|Fs],Rs,delay(St))).


%% poplist_queue(+Q_old:queue,?Els:list,-Q_new:queue) is nondet.
%
% Pops elements from Q_old onto Els to give Q_new. See pop_queue/3.
%
% Gives a.o. quick (and dirty?)
% ways of implementing queue-to-list 
% ( =|poplist_queue(Q_old,Els,Q_new), empty(Q_new)|= ) and  `take N'
% ( =|length(N,Els), poplist(Q_old,Els,Q_new)|= ). 
%
:- pred pops_queue(queue(T),list(T),queue(T)).
%
poplist_queue(Q,[],Q).
poplist_queue(Q,[El|Els],Q2):-
	pop_queue(Q,El,Q1),
	poplist_queue(Q1,Els,Q2).


%% injectlist_queue(+Els:list,+Q_old:queue,-Q_new:queue) is det.
%
% Injects elements of Els into Q_old to give Q_new. 
%
:- pred injects_queue(list(T),queue(T),queue(T)).
%
injectlist_queue([],Q,Q).
injectlist_queue([El|Els],Q,Q2):-
	inject_queue(El,Q,Q1),
	injectlist_queue(Els,Q1,Q2).


%% pushlist_queue(+Els:list,+Q_old:queue,-Q_new:queue) is det.
%
% Pushs elements of Els onto Q_old to give Q_new. 
%
:- pred pushs_queue(list(T),queue(T),queue(T)).
%
pushlist_queue([],Q,Q).
pushlist_queue([El|Els],Q,Q2):-
	push_queue(El,Q,Q1),
	pushlist_queue(Els,Q1,Q2).


% Aux predicates. See the paper for explanation. Compared to the
% paper, these are longer, `unrolled' versions that do not use the
% second-order predicate call/2.
%

% make_q/4      +State, +Fs     +Rs     -Q1
:- pred make_q(state(T),list(T),list(T),queue(T)).
%
make_q(ready,Fs,Rs,q(Fs1,[],St1)):-
	rot(Fs,Fs1,Rs,[[]|_],St1).
make_q(wait(A),Fs,Rs,q(Fs,Rs,St1)):-
    wait(A,St1).
make_q(delay(St1),Fs,Rs,q(Fs,Rs,St1)).         
make_q(rot(A1,A2,A3,A4),Fs,Rs,q(Fs,Rs,St1)):-
	rot(A1,A2,A3,A4,St1).


% wait/2     +Rs_rev -State
:- pred(wait(list(T),state(T))).
wait([],ready).
wait([_|Rs_rev],wait(Rs_rev)).


% rot/5     +Fs     ?Fs_tail +Rs     ?Rs_revs      -State
:- pred(rot(list(T),list(T), list(T),list(list(T)),state(T))).
%
rot([],[R|Rs_rev],[R],[Rs_rev],wait(Rs_rev)).
rot([F|Fs],[F|Fs_tail],[R|Rs],[Rs_rev|Rs_revs],rot(Fs,Fs_tail,Rs,Rs_revs)):-
	Rs_revs = [[R|Rs_rev]|_].
